import numpy as np

from drymass.cli import parse_funcs


def test_fbool():
    assert isinstance(parse_funcs.fbool("true"), bool)
    assert isinstance(parse_funcs.fbool("True"), bool)
    assert isinstance(parse_funcs.fbool("false"), bool)
    assert isinstance(parse_funcs.fbool("False"), bool)
    assert isinstance(parse_funcs.fbool("1"), bool)
    assert isinstance(parse_funcs.fbool("1."), bool)
    assert isinstance(parse_funcs.fbool("0"), bool)
    assert isinstance(parse_funcs.fbool("0."), bool)
    assert isinstance(parse_funcs.fbool(1.0), bool)
    assert isinstance(parse_funcs.fbool(1), bool)
    assert isinstance(parse_funcs.fbool(0.0), bool)
    assert isinstance(parse_funcs.fbool(0), bool)
    assert parse_funcs.fbool("true")
    assert parse_funcs.fbool("True")
    assert parse_funcs.fbool("1")
    assert parse_funcs.fbool("1.")
    assert parse_funcs.fbool(1.)
    assert parse_funcs.fbool(1)
    assert not parse_funcs.fbool("false")
    assert not parse_funcs.fbool("False")
    assert not parse_funcs.fbool("0.")
    assert not parse_funcs.fbool("0")
    assert not parse_funcs.fbool(0)
    assert not parse_funcs.fbool(0.)

    try:
        parse_funcs.fbool("")
    except ValueError:
        pass
    else:
        assert False, "emtpy string"


def test_fintlist():
    assert parse_funcs.fintlist("1, 2, 3") == [1, 2, 3]
    assert parse_funcs.fintlist("1.1, 2.1, 3.1") == [1, 2, 3]
    assert parse_funcs.fintlist([1, 2, 3]) == [1, 2, 3]
    assert parse_funcs.fintlist([1.1, 2.1, 3.1]) == [1, 2, 3]


def test_float01():
    assert parse_funcs.float01(.03) == .03

    try:
        parse_funcs.float01(1.03)
    except ValueError:
        pass
    else:
        assert False

    try:
        parse_funcs.float01(-1.03)
    except ValueError:
        pass
    else:
        assert False


def test_float_or_str():
    assert np.isnan(parse_funcs.float_or_str("nan"))
    assert isinstance(parse_funcs.float_or_str("1.1"), float)
    assert isinstance(parse_funcs.float_or_str(1), float)
    assert isinstance(parse_funcs.float_or_str(False), float)
    assert parse_funcs.float_or_str("1.1") == 1.1
    assert parse_funcs.float_or_str("1.1b") == "1.1b"
    assert parse_funcs.float_or_str("1.1e-5") == 1.1e-5


def test_float_tuple_or_one():
    assert parse_funcs.floattuple_or_one("1") == 1
    assert parse_funcs.floattuple_or_one("+1") == 1
    assert parse_funcs.floattuple_or_one("-1") == -1
    assert parse_funcs.floattuple_or_one(1) == 1
    assert parse_funcs.floattuple_or_one(-1) == -1
    assert parse_funcs.floattuple_or_one(1.0) == 1
    assert parse_funcs.floattuple_or_one(-1.0) == -1
    assert np.allclose(parse_funcs.floattuple_or_one([1, 2]), [1, 2])
    assert np.allclose(parse_funcs.floattuple_or_one(["1", "2"]), [1, 2])
    assert np.allclose(parse_funcs.floattuple_or_one("1.0, 2"), [1, 2])
    assert np.allclose(parse_funcs.floattuple_or_one((1, 2)), [1, 2])
    assert np.allclose(parse_funcs.floattuple_or_one([1.2, 2.9]), [1.2, 2.9])
    assert np.allclose(parse_funcs.floattuple_or_one(np.array([1, 2])), [1, 2])

    try:
        parse_funcs.floattuple_or_one("1,2,3,4")
    except ValueError:
        pass
    else:
        assert False

    try:
        parse_funcs.floattuple_or_one("nan")
    except ValueError:
        pass
    else:
        assert False

    try:
        parse_funcs.floattuple_or_one(2)
    except ValueError:
        pass
    else:
        assert False


def test_int_or_path():
    assert isinstance(parse_funcs.int_or_path("1.1"), int)
    assert isinstance(parse_funcs.int_or_path(1), int)
    assert isinstance(parse_funcs.int_or_path(False), int)
    assert parse_funcs.int_or_path("1.1") == 1
    assert parse_funcs.int_or_path("1.1b") == "1.1b"
    assert parse_funcs.int_or_path("2") == 2


def test_lcstr():
    assert parse_funcs.lcstr("ASD") == "asd"


def test_strlist_vsort():
    assert parse_funcs.strlist_vsort("1.1, 2.1, 3") == ["1.1", "2.1", "3"]
    assert parse_funcs.strlist_vsort("1.10, 1.9") == ["1.9", "1.10"]


def test_tupletupleint():
    assert parse_funcs.tupletupleint("[(1, 2), (3, 4)]") == ((1, 2), (3, 4))
    assert parse_funcs.tupletupleint("1, 2, 3, 4") == ((1, 2), (3, 4))
    assert parse_funcs.tupletupleint("1.1, 2.1, 3.1, 4.1") == ((1, 2), (3, 4))
    assert parse_funcs.tupletupleint([(1, 2), (3, 4)]) == ((1, 2), (3, 4))
    assert parse_funcs.tupletupleint([(1.1, 2.), (3.0, 4)]) == ((1, 2), (3, 4))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import multiprocessing as mp
import sys
import time
from threading import Thread


def print_percentage(print_prefix, count, max_count, abort):
    """Track count and max_count and print percentage on command line"""
    while max_count.value == 0 or count.value != max_count.value:
        if max_count.value == 0:
            perc = 0
        else:
            perc = count.value/max_count.value * 100
        print("{}{:.1f}%".format(print_prefix, perc), end="\r", flush=True)
        time.sleep(0.1)
        if abort.value:
            break
    # cleanup the console
    print(" "*(6+len(print_prefix)), end="\r")
    print(print_prefix, end="", flush=True)


class TaskWatcher(object):
    def __init__(self, print_prefix="Performing task..."):
        """A progress watcher for the command line"""
        self.print_prefix = print_prefix
        self.count = mp.Value('I', 0, lock=True)
        self.max_count = mp.Value('I', 0, lock=True)
        self.abort = mp.Value('I', 0, lock=True)
        self.thread = Thread(target=print_percentage,
                             args=(print_prefix, self.count, self.max_count,
                                   self.abort))
        # causes the thread to terminate when the main process ends
        self.thread.daemon = True

    def __enter__(self):
        self.thread.start()
        return self

    def __exit__(self, type, value, traceback):
        # check whether an exception has been raised
        if sys.exc_info() != (None, None, None):
            # A traceback will be printed; Keep the last line of the
            # progress visible.
            print("", flush=True)
        else:
            self.abort.value += 1  # tell the printer to stop doing its thing
            self.thread.join(timeout=.2)  # be patient
            if (self.max_count.value != 0
                    and self.max_count.value != self.count.value):
                # The tracked function is not properly updating
                # count and max_count.
                raise ValueError(
                    "`count`={} did not count to `max_count`={}".format(
                        self.count.value, self.max_count.value))



import io
from os import fspath
import sys

import matplotlib.image as mpimg
import qpimage
import tifffile

from ..extractroi import extract_roi

from . import config
from .converting import cli_convert
from . import dialog
from . import plot
from .task_watcher import TaskWatcher


#: Matplotlib images of sensor data with labeled ROI (TIFF)
FILE_SENSOR_WITH_ROI_IMAGE = "sensor_roi_images.tif"


def cli_extract_roi(path=None, ret_data=False, profile=None):
    """Extract regions of interest"""
    description = "Extract reqions of interest in quantitative phase" \
                  + "microscopy data for further analysis in DryMass."
    path_in, path_out = dialog.main(path=path,
                                    description=description,
                                    profile=profile)
    if isinstance(path_in, list):
        # recursive analysis
        for ii, pi in enumerate(path_in):
            print("Analyzing dataset {}/{}.".format(ii+1, len(path_in)))
            cli_extract_roi(path=pi)
        # nothing else to do
        return
    # cli_convert will ask for the required meta data
    h5series = cli_convert(path=path_in, ret_data=True)
    # get the configuration after cli_convert was run
    cfg = config.ConfigFile(path_out)
    # background correction
    if cfg["bg"]["enabled"]:
        bg_amp_kw = {"fit_offset": cfg["bg"]["amplitude offset"],
                     "fit_profile": cfg["bg"]["amplitude profile"],
                     "border_perc": cfg["bg"]["amplitude border perc"],
                     "border_px": cfg["bg"]["amplitude border px"],
                     }
        bg_pha_kw = {"fit_offset": cfg["bg"]["phase offset"],
                     "fit_profile": cfg["bg"]["phase profile"],
                     "border_perc": cfg["bg"]["phase border perc"],
                     "border_px": cfg["bg"]["phase border px"],
                     }
        # Only get the Canny edge detection parameters if needed
        if not (cfg["bg"]["phase mask sphere"] is None
                and cfg["bg"]["amplitude mask sphere"] is None):
            dialog.main(path=path, req_meta=["medium index"])
            # This will generate the [sphere] section in drymass.cfg
            edge_kw = {
                "clip_rmin": cfg["sphere"]["edge clip radius min"],
                "clip_rmax": cfg["sphere"]["edge clip radius max"],
                "mult_coarse": cfg["sphere"]["edge coarse"],
                "mult_fine": cfg["sphere"]["edge fine"],
                "maxiter": cfg["sphere"]["edge iter"],
            }
        else:
            edge_kw = {}
    else:
        bg_amp_kw = None
        bg_pha_kw = None
        edge_kw = {}

    with TaskWatcher("Extracting ROIs... ") as tw:
        h5roi, rmgr, changed = extract_roi(
            h5series=h5series,
            dir_out=path_out,
            size_m=cfg["specimen"]["size um"] * 1e-6,
            size_var=cfg["roi"]["size variation"],
            max_ecc=cfg["roi"]["eccentricity max"],
            dist_border=cfg["roi"]["dist border px"],
            pad_border=cfg["roi"]["pad border px"],
            exclude_overlap=cfg["roi"]["exclude overlap px"],
            threshold=cfg["roi"]["threshold"],
            ignore_data=cfg["roi"]["ignore data"],
            force_roi=cfg["roi"]["force"],
            bg_amp_kw=bg_amp_kw,
            bg_amp_bin=cfg["bg"]["amplitude binary threshold"],
            bg_amp_mask_radial_clearance=cfg["bg"]["amplitude mask sphere"],
            bg_pha_kw=bg_pha_kw,
            bg_pha_bin=cfg["bg"]["phase binary threshold"],
            bg_pha_mask_radial_clearance=cfg["bg"]["phase mask sphere"],
            bg_sphere_edge_kw=edge_kw,
            search_enabled=cfg["roi"]["enabled"],
            ret_roimgr=True,
            ret_changed=True,
            count=tw.count,
            max_count=tw.max_count,
        )
    if changed:
        print("Done.")
    else:
        print("Reusing.")
    if len(rmgr) == 0:
        print("No ROIs could be found!\n"
              + "Please try editing '{}':\n".format(cfg.path)
              + "- correct specimen size "
              + "({})\n".format(strpar(cfg, "specimen", "size um"))
              + "- increase allowed size variation "
              + "({})\n".format(strpar(cfg, "roi", "size variation"))
              + "- increase maximum allowed eccentricity "
              + "({})\n".format(strpar(cfg, "roi", "eccentricity max"))
              + "- reduce minimum distance to image border in pixels "
              + "({})\n".format(strpar(cfg, "roi", "dist border px"))
              + "- reduce minimum distance inbetween ROIs "
              + "({})".format(strpar(cfg, "roi", "exclude overlap px"))
              )
        sys.exit(1)

    if changed and cfg["output"]["roi images"]:
        print("Plotting detected ROIs... ", end="", flush=True)
        tifout = path_out / FILE_SENSOR_WITH_ROI_IMAGE
        # plot h5series and rmgr with matplotlib
        with qpimage.QPSeries(h5file=h5series, h5mode="r") as qps, \
                tifffile.TiffWriter(fspath(tifout), imagej=True) as tf:
            for ii in range(len(qps)):
                # new indexing convention in drymass 0.6.0
                image_index = ii + 1
                rois = rmgr.get_from_image_index(image_index)
                imio = io.BytesIO()
                plot.plot_qpi_phase(qps[ii],
                                    rois=rois,
                                    path=imio,
                                    labels_excluded=cfg["roi"]["ignore data"])
                imio.seek(0)
                imdat = (mpimg.imread(imio) * 255).astype("uint8")
                tf.save(imdat, compress=9)
        print("Done")

    if ret_data:
        return h5roi


def strpar(cfg, section, key):
    """String representation of a section/key combination"""
    return "[{}] {} = {}".format(section, key, cfg[section][key])



import argparse
import pathlib
import tempfile
import warnings

import numpy as np
import qpimage

from drymass.cli import cli_convert, config, dialog, profile


def setup_config(pxsize=1e-6, medium_index=1.335, wavelength=550e-9):
    _, path = tempfile.mkstemp(prefix="drymass_test_config_", suffix=".cfg")
    cfg = config.ConfigFile(path=path)
    cfg.set_value("bg", "phase profile", "tilt")
    cfg.set_value("bg", "amplitude profile", "tilt")
    cfg.set_value("roi", "dist border px", 3)
    cfg.set_value("roi", "exclude overlap px", 5)
    cfg.set_value("roi", "pad border px", 7)
    cfg.set_value(section="meta", key="pixel size um", value=pxsize*1e6)
    cfg.set_value(section="meta", key="wavelength nm",
                  value=wavelength*1e9)
    cfg.set_value(section="meta", key="medium index", value=medium_index)

    return cfg.path


def setup_test_data(radius_px=30, size=200, num=1):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp), which_data="phase,amplitude")
    path_in = tempfile.mktemp(suffix=".h5", prefix="drymass_test_cli_profile")
    path_in = pathlib.Path(path_in)
    with qpimage.QPSeries(h5file=path_in, h5mode="w", identifier="tes") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="test_{}".format(ii))
    path_out = path_in.with_name(path_in.name + dialog.OUTPUT_SUFFIX)
    path_out.mkdir()
    return qpi, path_in, path_out


def test_add_fail():
    path = setup_config()
    argsadd = argparse.Namespace(subparser_name="add",
                                 name="test_8473_prof",
                                 path=path)
    profile.cli_profile(args=argsadd)
    try:
        profile.cli_profile(args=argsadd)
    except OSError:
        pass
    else:
        assert False
    # remove the profile again
    argsrem = argparse.Namespace(subparser_name="remove",
                                 name="test_8473_prof")
    profile.cli_profile(args=argsrem)


def test_add_remove():
    path = setup_config()
    argsadd = argparse.Namespace(subparser_name="add",
                                 name="test_8472_prof",
                                 path=path)
    profile.cli_profile(args=argsadd)
    # verify that the profile was imported
    pps = profile.get_profile_path(name="test_8472_prof")
    assert pps.exists()
    # remove the profile again
    argsrem = argparse.Namespace(subparser_name="remove",
                                 name="test_8472_prof")
    profile.cli_profile(args=argsrem)
    assert not pps.exists()


def test_convert_with_profile():
    cfgpath = setup_config(pxsize=1.34e-6,
                           medium_index=1.346,
                           wavelength=554.2e-9)
    _, path_in, path_out = setup_test_data()
    argsadd = argparse.Namespace(subparser_name="add",
                                 name="test_8440_prof_convert",
                                 path=cfgpath)
    profile.cli_profile(args=argsadd)
    # perform conversion
    h5data = cli_convert(path=path_in,
                         ret_data=True,
                         profile="test_8440_prof_convert")
    cfg = config.ConfigFile(path_out)
    assert np.allclose(cfg["meta"]["medium index"], 1.346)
    assert np.allclose(cfg["meta"]["pixel size um"], 1.34)
    assert np.allclose(cfg["meta"]["wavelength nm"], 554.2)

    with qpimage.QPSeries(h5file=h5data, h5mode="r") as qps:
        assert np.allclose(qps[0]["medium index"], 1.346)
        assert np.allclose(qps[0]["pixel size"], 1.34e-6)
        assert np.allclose(qps[0]["wavelength"], 554.2e-9)

    # cleanup
    argsrem = argparse.Namespace(subparser_name="remove",
                                 name="test_8440_prof_convert")
    profile.cli_profile(args=argsrem)


def test_export():
    path = setup_config()
    argsadd = argparse.Namespace(subparser_name="add",
                                 name="test_8491_prof",
                                 path=path)
    profile.cli_profile(args=argsadd)
    # export
    tdir = tempfile.mkdtemp(prefix="test_drymass_profile_export_")
    argsexp = argparse.Namespace(subparser_name="export",
                                 path=tdir)
    profile.cli_profile(args=argsexp)
    assert (pathlib.Path(tdir) / "profile_test_8491_prof.cfg").exists()
    # cleanup
    argsrem = argparse.Namespace(subparser_name="remove",
                                 name="test_8491_prof")
    profile.cli_profile(args=argsrem)


def test_get_profile_path():
    path = setup_config()
    assert path == profile.get_profile_path(name=path)


def test_list_none(capsys):
    if profile.get_profiles():
        warnings.warn("Test cannot succeed, b/c there are user profiles.")
    else:
        argslist = argparse.Namespace(subparser_name="list")
        profile.cli_profile(args=argslist)
        captured = capsys.readouterr()
        assert captured.out.strip() == "No profiles in local library."


def test_list_profile(capsys):
    path = setup_config()
    argsadd = argparse.Namespace(subparser_name="add",
                                 name="test_8490_prof",
                                 path=path)
    profile.cli_profile(args=argsadd)
    argslist = argparse.Namespace(subparser_name="list")
    profile.cli_profile(args=argslist)
    captured = capsys.readouterr()
    assert "- test_8490_prof:" in captured.out.strip()
    argsrem = argparse.Namespace(subparser_name="remove",
                                 name="test_8490_prof")
    profile.cli_profile(args=argsrem)


def test_remove_fail():
    argsrem = argparse.Namespace(subparser_name="remove",
                                 name="test_8474_prof")
    try:
        profile.cli_profile(args=argsrem)
    except OSError:
        pass
    else:
        assert False


if __name__ == "__main__":
    # Run all tests
    print("Cannot run all tests b/c of usage of `capsys` fixture! "
          "Please use py.test.")



import pathlib

from ..converter import convert

from . import config
from . import dialog
from .task_watcher import TaskWatcher


def cli_convert(path=None, ret_data=False, profile=None):
    """Convert input data to QPSeries data"""
    description = "Convert raw quantitative phase microscopy data to " \
                  + "the qpimage file format for further analysis in " \
                  + "DryMass."
    path_in, path_out = dialog.main(path=path,
                                    req_meta=["pixel size um",
                                              "wavelength nm"],
                                    description=description,
                                    profile=profile)
    if isinstance(path_in, list):
        # recursive analysis
        for ii, pi in enumerate(path_in):
            print("Analyzing dataset {}/{}.".format(ii+1, len(path_in)))
            cli_convert(path=pi)
        # nothing else to do
        return
    cfg = config.ConfigFile(path_out)

    meta_data = {"pixel size": cfg["meta"]["pixel size um"] * 1e-6,
                 "wavelength": cfg["meta"]["wavelength nm"] * 1e-9,
                 "medium index": cfg["meta"]["medium index"],
                 }

    holo_kw = {"filter_name": cfg["holo"]["filter name"],
               "filter_size": cfg["holo"]["filter size"],
               "sideband": cfg["holo"]["sideband"],
               }

    bg_data_amp = parse_bg_value(cfg["bg"]["amplitude data"],
                                 reldir=path_in.parent)
    bg_data_pha = parse_bg_value(cfg["bg"]["phase data"],
                                 reldir=path_in.parent)

    with TaskWatcher("Converting input data... ") as tw:
        h5series, ds, changed = convert(
            path_in=path_in,
            dir_out=path_out,
            meta_data=meta_data,
            holo_kw=holo_kw,
            bg_data_amp=bg_data_amp,
            bg_data_pha=bg_data_pha,
            write_tif=cfg["output"]["sensor tif data"],
            ret_dataset=True,
            ret_changed=True,
            count=tw.count,
            max_count=tw.max_count,
            )
    if "hologram" not in ds.storage_type:
        # remove "holo" section from configuration
        cfg.remove_section("holo")
    if changed:
        print("Done.")
    else:
        print("Reusing.")
    if ret_data:
        return h5series


def parse_bg_value(bg, reldir):
    """Determine the background to use from the configuration key"""
    if isinstance(bg, str):
        # can be a file name relative to the input directory
        # or an absolute path.
        reldir = pathlib.Path(reldir)
        path = reldir / bg
        if not path.exists():
            path = pathlib.Path(bg)
        bg = path
    return bg



import numpy as np

from drymass import threshold as thr


def test_basic():
    size = 200
    image = np.zeros((size, size), dtype=float)
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    radius = 30
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    image[r < radius] = 1.3
    assert np.all(thr.image2mask(image, 1) == (r < radius))
    assert np.all(thr.image2mask(image, "dm-nuclei") == (r < radius))


def test_invert():
    size = 200
    image = np.zeros((size, size), dtype=float)
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    radius = 30
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    image[r < radius] = 1.3
    normal = thr.image2mask(image, 1)
    invert = thr.image2mask(image, 1, invert=True)
    assert np.all(normal == ~invert)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib

import numpy as np
import qpimage
import qpsphere

from . import util

#: Output sphere analysis qpimage.QPSeries data
FILE_SPHERE_DATA = "sphere_{}_{}_data.h5"
#: Output sphere analysis statistics
FILE_SPHERE_STAT = "sphere_{}_{}_statistics.txt"


class EdgeDetectionFailedWarning(UserWarning):
    pass


def analyze_sphere(h5roi, dir_out, r0=10e-6, method="edge",
                   model="projection", edgekw={}, imagekw={},
                   alpha=.18, rad_fact=1.2, ret_changed=False,
                   ret_reused=False, count=None, max_count=None):
    """Perform sphere analysis

    Parameters
    ----------
    h5series: str
        Path of qpimage.QPSeries hdf5 file
    dir_out: str
        Path to output directory
    r0: float
        Initial radius
    method: str
        Either "edge" or "image"; see :ref:`config_sphere` for
        more information.
    model: str
        Propagation model to use; see :ref:`config_sphere` for
        more information.
    edgekw: dict
        Keyword arguments to :func:`qpsphere.edgefit.contour_canny`
    imagekw: dict
        Keyword arguments to :func:`qpsphere.imagefit.alg.match_phase`
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Radial inclusion factor for dry mass computation
    ret_changed: bool
        Return boolean indicating whether the sphere data on disk was
        created/updated (True) or whether only previously created ROI
        data was used (False).
    ret_reused: bool
        Return integer indicating how many previous fits
        were reused.
    count, max_count: multiprocessing.Value
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.
    """
    dir_out = pathlib.Path(dir_out).resolve()

    h5out = dir_out / FILE_SPHERE_DATA.format(method, model)
    statout = dir_out / FILE_SPHERE_STAT.format(method, model)

    with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qps:
        dataid, roiparid, roiexclid = qps.identifier.split(":")
        cfgid = util.hash_object([r0,
                                  method,
                                  model,
                                  edgekw,
                                  imagekw if method == "image" else None,
                                  alpha,
                                  rad_fact])
    # Previous reference dataset may contain valuable fitting results
    h5ref = None
    changed = True
    reused = 0
    if util.is_series_file(h5out):
        with qpimage.QPSeries(h5file=h5out, h5mode="r") as qps_ref:
            refids = qps_ref.identifier.split(":")
            refids.pop(2)  # remove roiexclid from identifier
        if [dataid, roiparid, cfgid] == refids:
            # reuse (rename to temporary file)
            h5ref = h5out.with_suffix(".ref.h5")
            h5out.rename(h5ref)
            changed = False

    ids_ref = []
    if h5ref is not None:
        with qpimage.QPSeries(h5file=h5ref, h5mode="r") as qps_ref:
            ids_ref = [qpi["identifier"] for qpi in qps_ref]

    # initialize output file with identifier
    identifier = ":".join([dataid, roiparid, roiexclid, cfgid])
    with qpimage.QPSeries(h5file=h5out, h5mode="w",
                          identifier=identifier) as qps_out:
        pass

    with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qps_in, \
            statout.open(mode="w") as fd:

        header = ["identifier",
                  "index",
                  "radius_um",
                  "rel_dry_mass_pg",
                  "abs_dry_mass_pg",
                  "time",
                  "medium",
                  ]
        fd.write("#" + "\t".join(header) + "\r\n")

        if max_count is not None:
            with max_count.get_lock():
                max_count.value += len(qps_in)

        for qpi in qps_in:
            simident = "{}:{}".format(qpi["identifier"], model)
            if simident in ids_ref:
                with qpimage.QPSeries(h5file=h5ref, h5mode="r") as qps_ref:
                    qpi_sim = qps_ref[simident].copy()
                n = qpi_sim["sim index"]
                r = qpi_sim["sim radius"]
                c = qpi_sim["sim center"]
                ids_ref.remove(simident)
                reused += 1
            else:
                try:
                    # fit sphere model
                    n, r, c, qpi_sim = qpsphere.analyze(qpi,
                                                        r0=r0,
                                                        method=method,
                                                        model=model,
                                                        edgekw=edgekw,
                                                        imagekw=imagekw,
                                                        ret_center=True,
                                                        ret_qpi=True)
                except qpsphere.models.excpt.UnsupportedModelParametersError:
                    print("Skipping object {} ".format(qpi["identifier"])
                          + "because unsupported model parameters were "
                          + "encountered.")
                    continue
                except BaseException as exc:
                    # Be more verbose
                    exc.args = ("ROI {}: ".format(qpi["identifier"])
                                + exc.args[0],)
                    raise
                else:
                    changed = True
            # write simulation results
            with qpimage.QPSeries(h5file=h5out, h5mode="a") as qps_out:
                qps_out.add_qpimage(qpi=qpi_sim, identifier=simident)
            # finally, update text file
            if "time" in qpi:
                qptime = qpi["time"]
            else:
                qptime = np.nan
            data = {
                "identifier": qpi["identifier"],
                "index": n,
                "radius_um": r * 1e6,
                "abs_dry_mass_pg": absolute_dry_mass_sphere(
                    qpi=qpi,
                    radius=r,
                    center=c,
                    alpha=alpha,
                    rad_fact=rad_fact
                ) * 1e12,
                "rel_dry_mass_pg": relative_dry_mass(
                    qpi=qpi,
                    radius=r,
                    center=c,
                    alpha=alpha,
                    rad_fact=rad_fact
                ) * 1e12,
                "time": qptime,
                "medium": qpi["medium index"]
            }
            fd.write("\t".join([str(data[k]) for k in header]) + "\r\n")
            fd.flush()
            if count is not None:
                with count.get_lock():
                    count.value += 1

    if ids_ref:
        # leftovers
        changed = True
    # cleanup
    if h5ref is not None:
        h5ref.unlink()

    ret = [h5out]
    if ret_changed:
        ret.append(changed)
    if ret_reused:
        ret.append(reused)

    if len(ret) == 1:
        ret = ret[0]
    return ret


def absolute_dry_mass_sphere(qpi, radius, center, alpha=.18, rad_fact=1.2):
    """Compute absolute dry mass of a spherical phase object

    The absolute dry mass is computed with

    .. code::

        m_abs = m_rel + m_sup
        m_rel = lambda / (2*PI*alpha) * phi_tot * deltaA
        m_sup = 4*PI / (3*alpha) * radius^3 (n_med - n_PBS)

    with the vacuum wavelength ``lambda``, the total phase
    retardation in the area of interest ``phi_tot``, the pixel
    area ``deltaA``, the refractive index of the medium `n_med`
    (stored in ``qpi.meta``), and the refractive index of phosphate
    buffered saline (PBS) ``n_PBS=1.335``.

    This is the *absolute* dry mass, because it takes into account
    the offset caused by the suppressed density in the phase data.

    Parameters
    ----------
    qpi: qpimage.QPImage
        QPI data
    center: tuble (x,y)
        Center of the sphere [px]
    radius: float
        Radius of the sphere [m]
    wavelength: float
        The wavelength of the light [m]
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Inclusion factor that scales `radius` to increase
        the area used for phase summation; if the backgound
        phase exhibits a noise phase signal, positive and
        negative contributions cancel out and `rad_fact`
        does not have an effect above a certain critical value.

    Returns
    -------
    dry_mass: float
        The absolute dry mass of the sphere [g]
    """
    dm_rel = relative_dry_mass(qpi=qpi,
                               radius=radius,
                               center=center,
                               alpha=alpha,
                               rad_fact=rad_fact)
    medium_index = qpi["medium index"]
    dm_sup = 4 / 3 * np.pi / (alpha * 1e-6) * \
        radius**3 * (medium_index - 1.335)
    return dm_rel + dm_sup


def relative_dry_mass(qpi, radius, center, alpha=.18, rad_fact=1.2):
    """Compute relative dry mass of a circular area in QPI

    The dry mass is computed with

    m_rel = lambda / (2*PI*alpha) * phi_tot * deltaA

    with the vacuum wavelength `lambda`, the total phase
    retardation in the area of interest `phi_tot`, and the pixel
    area `deltaA`.

    This is the *relative* dry mass, because it is computed relative
    to the surrounding medium (phi_tot) and not relative to water.

    Parameters
    ----------
    qpi: qpimage.QPImage
        QPI data
    center: tuble (x,y)
        Center of the area of interest [px]
    radius: float
        Radius of the area of interest [m]
    wavelength: float
        The wavelength of the light [m]
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Inclusion factor that scales `radius` to increase
        the area used for phase summation; if the backgound
        phase exhibits a noise phase signal, positive and
        negative contributions cancel out and `rad_fact`
        does not have an effect above a certain critical value.

    Returns
    -------
    dry_mass: float
        The relative dry mass of the object [g]
    """
    image = qpi.pha
    rincl = radius / qpi["pixel size"] * rad_fact
    wavelength = qpi["wavelength"]
    sx, sy = qpi.shape
    x = np.arange(sx).reshape(-1, 1)
    y = np.arange(sy).reshape(1, -1)
    discsq = ((x - center[0])**2 + (y - center[1])**2)
    area = discsq < rincl**2
    phi_tot = np.sum(image[area])
    # compute dry mass
    pxarea = qpi["pixel size"]**2
    # convert alpha mL/g to m³/g
    fact = 1e-6
    # [kg]
    dm = wavelength / (2 * np.pi * alpha * fact) * phi_tot * pxarea
    return dm
